# task1 = task1 or """
# 1. read through the given functions, and try to understand what they do.
# 2. if any function appear undefined in the environment, define them according to given information
# 3. add a few test cases and check correctness
# """

# context = "codes to evaluate: "+ context

# task2 = task2 or f"""
# 1. validate correctness, and add 5-10 new functions using pure python
# 2. improve time/space complexity while preserving usage behavior
# 3. run to make sure correctness
# """

# task3 = task3 or f"""
# present the full implementation and save as: {filename}.txt
# 1. add typing and google format docstrings
# 2. add a couple examples in the docstrings
# 3. add unittests
# """
