base_lion_fields = [
    "ln_id",
    "timestamp",
    "metadata",
    "extra_fields",
    "content",
    "created",
    "embedding",
]

lc_meta_fields = ["lc", "type", "id", "langchain", "lc_type", "lc_id"]

llama_meta_fields = [
    "id_",
    "embedding",
    "excluded_embed_metadata_keys",
    "excluded_llm_metadata_keys",
    "relationships",
    "start_char_idx",
    "end_char_idx",
    "class_name",
    "text_template",
    "metadata_template",
    "metadata_seperator",
]

SYSTEM_FIELDS = [
    "ln_id",
    "timestamp",
    "metadata",
    "meta",
    "extra_fields",
    "content",
    "created",
    "form",
    "report",
    "work",
    "assignment",
    "assignments",
    "input_fields",
    "requested_fields",
    "instruction",
    "system",
    "strict",
]
